@@ -22,20 +22,22 @@
 /***************************************************************************/
 
 /*
-   1999/08/13 implementation of FtpClose2() and FtpChmod() 
+   1999/08/13 implementation of FtpClose2() and FtpChmod()
               by Masashi Itou <webmaster@maitou.gr.jp>
    1999/11/04 implementation of a progress information in FtpXfer()
-              by Yuuki NINOMIYA <gm@debian.or.jp>   
+              by Yuuki NINOMIYA <gm@debian.or.jp>
    1999/??/?? commenting out of calling perror()
-              by Yuuki NINOMIYA <gm@debian.or.jp>   
+              by Yuuki NINOMIYA <gm@debian.or.jp>
    2000/01/09 FTP mode is decided by ftp_passive[host_number]
               instead of FTPLIB_DEFMODE in FtpConnect()
-              by Yuuki NINOMIYA <gm@debian.or.jp>   
+              by Yuuki NINOMIYA <gm@debian.or.jp>
+   2001/09/05 Added support for SSL
+              by John E. Davis <jed@jedsoft.org>
 */
-	
+
 #ifdef HAVE_CONFIG_H
 #  include "config.h"
-#endif  
+#endif
 
 #include "intl.h"
 #include "variable.h"
@@ -69,6 +71,17 @@
 #include <winsock.h>
 #endif
 
+#ifdef USE_SSL
+# include <openssl/ssl.h>
+# include <openssl/err.h>
+# include <openssl/rand.h>
+
+static int print_ssl_err_message (char *, SSL *, int);
+static int authorize_ssl (netbuf *);
+static int initialize_ssl (void);
+static int setup_data_connection_ssl (netbuf *);
+#endif
+
 #define BUILDING_LIBRARY
 #include "ftplib.h"
 
@@ -93,7 +106,11 @@
 
 struct NetBuf {
     char *cput,*cget;
+    char *hostname;
     int handle;
+#ifdef USE_SSL
+    SSL *ssl_con;
+#endif
     int cavail,cleft;
     char *buf;
     int dir;
@@ -113,7 +130,7 @@ struct NetBuf {
  * But the following version and copyright strings live in the binary.
  * Please type: `% strings weex | grep ftplib' :-)
  *
- * 1999/11/06 Yuuki NINOMIYA <gm@debian.or.jp>   
+ * 1999/11/06 Yuuki NINOMIYA <gm@debian.or.jp>
  */
 char *version =
     "ftplib Release 3.1 6/xx/98, copyright 1996, 1997, 1998 Thomas Pfau";
@@ -129,6 +146,7 @@ GLOBALDEF int ftplib_debug = 0;
 #define net_write(x,y,z) send(x,y,z,0)
 #define net_close closesocket
 #endif
+
 
 #if defined(NEED_MEMCCPY)
 /*
@@ -205,6 +223,43 @@ static int socket_wait(netbuf *ctl)
     return rv;
 }
 
+static int netbuf_read (netbuf *c, char *buf, int n)
+{
+#ifdef USE_SSL
+   if (c->ssl_con != NULL)
+     {
+	int status = SSL_read (c->ssl_con, buf, n);
+	if (status > 0)
+	  return status;
+
+	/* Sometimes a data connection will just close at EOF */
+	if (status == 0)
+	  return -1;
+
+	fprintf (stderr, "SSL_read error on %d\n", c->handle);
+	return print_ssl_err_message ("SSL_read", c->ssl_con, status);
+     }
+#endif
+
+   return net_read (c->handle, buf, n);
+}
+
+static int netbuf_write (netbuf *c, char *buf, unsigned int buflen)
+{
+#ifdef USE_SSL
+   if (c->ssl_con != NULL)
+     {
+	int status = SSL_write (c->ssl_con, buf, buflen);
+	if (status > 0)
+	  return status;
+
+	return print_ssl_err_message ("SSL_write", c->ssl_con, status);
+     }
+   else
+#endif
+     return net_write (c->handle, buf, buflen);
+}
+
 /*
  * read a line of text
  *
@@ -265,7 +320,7 @@ static int readline(char *buf,int max,netbuf *ctl)
 	}
 	if (!socket_wait(ctl))
 	    return retval;
-    	if ((x = net_read(ctl->handle,ctl->cput,ctl->cleft)) == -1)
+    	if ((x = netbuf_read(ctl,ctl->cput,ctl->cleft)) == -1)
     	{
 	    /* perror("read"); */
 	    retval = -1;
@@ -303,7 +358,7 @@ static int writeline(char *buf, int len, netbuf *nData)
 	    {
 		if (!socket_wait(nData))
 		    return x;
-		w = net_write(nData->handle, nbp, FTPLIB_BUFSIZ);
+		w = netbuf_write(nData, nbp, FTPLIB_BUFSIZ);
 		if (w != FTPLIB_BUFSIZ)
 		{
 		    printf("net_write(1) returned %d, errno = %d\n", w, errno);
@@ -317,7 +372,7 @@ static int writeline(char *buf, int len, netbuf *nData)
 	{
 	    if (!socket_wait(nData))
 		return x;
-	    w = net_write(nData->handle, nbp, FTPLIB_BUFSIZ);
+	    w = netbuf_write(nData, nbp, FTPLIB_BUFSIZ);
 	    if (w != FTPLIB_BUFSIZ)
 	    {
 		printf("net_write(2) returned %d, errno = %d\n", w, errno);
@@ -331,7 +386,7 @@ static int writeline(char *buf, int len, netbuf *nData)
     {
 	if (!socket_wait(nData))
 	    return x;
-	w = net_write(nData->handle, nbp, nb);
+	w = netbuf_write(nData, nbp, nb);
 	if (w != nb)
 	{
 	    printf("net_write(3) returned %d, errno = %d\n", w, errno);
@@ -352,9 +407,10 @@ static int readresp(char c, netbuf *nControl)
     char match[5];
     if (readline(nControl->response,256,nControl) == -1)
     {
-	/* perror("Control socket read failed"); */
-	return 0;
+       perror("Control socket read failed");
+       return 0;
     }
+
     if (ftplib_debug > 1)
 	fprintf(stderr,"%s",nControl->response);
     if (nControl->response[3] == '-')
@@ -376,6 +432,18 @@ static int readresp(char c, netbuf *nControl)
     }
     if (nControl->response[0] == c)
 	return 1;
+#ifdef USE_SSL
+   /* ftp-ssl in debian is reported to be somewhat inconsistent with other
+    * versions of ftp-ssl.  In particular, the return code for the
+    * AUTH SSL appears to vary.  So, support them both with this hack.
+    */
+   if (c == 127)
+     {
+	c = nControl->response[0];
+	if ((c == '2') || (c == '3'))
+	  return 1;
+     }
+#endif
     return 0;
 }
 
@@ -409,7 +477,7 @@ GLOBALDEF char *FtpLastResponse(netbuf *nControl)
  *
  * return 1 if connected, 0 if not
  */
-GLOBALDEF int FtpConnect(const char *host, netbuf **nControl)
+GLOBALDEF int FtpConnect(const char *host, netbuf **nControl, int use_ssl)
 {
     int sControl;
     struct sockaddr_in sin;
@@ -433,12 +501,12 @@ GLOBALDEF int FtpConnect(const char *host, netbuf **nControl)
 #if defined(VMS)
     	sin.sin_port = htons(21);
 #else
-    	if ((pse = getservbyname("ftp","tcp")) == NULL)
-    	{
+        if ((pse = getservbyname("ftp","tcp")) == NULL)
+	{
 	    /* perror("getservbyname"); */
 	    return 0;
-    	}
-    	sin.sin_port = pse->s_port;
+	}
+        sin.sin_port = pse->s_port;
 #endif
     }
     else
@@ -465,7 +533,6 @@ GLOBALDEF int FtpConnect(const char *host, netbuf **nControl)
     	}
     	memcpy((char *)&sin.sin_addr, phe->h_addr, phe->h_length);
     }
-    free(lhost);
     sControl = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
     if (sControl == -1)
     {
@@ -514,7 +581,12 @@ GLOBALDEF int FtpConnect(const char *host, netbuf **nControl)
     ctrl->xfered = 0;
     ctrl->xfered1 = 0;
     ctrl->cbbytes = 0;
-    if (readresp('2', ctrl) == 0)
+    ctrl->hostname = lhost;
+    if ((readresp('2', ctrl) == 0)
+#ifdef USE_SSL
+	|| (use_ssl && (-1 == authorize_ssl (ctrl)))
+#endif
+	)
     {
 	net_close(sControl);
 	free(ctrl->buf);
@@ -572,15 +644,15 @@ GLOBALDEF int FtpOptions(int opt, long val, netbuf *nControl)
  */
 static int FtpSendCmd(const char *cmd, char expresp, netbuf *nControl)
 {
-    char buf[256];
+    char buf[4096];		       /* large buffer for SSL */
     if (nControl->dir != FTPLIB_CONTROL)
 	return 0;
     if (ftplib_debug > 2)
-	fprintf(stderr,"%s\n",cmd);
+     fprintf(stderr," -->%s\n",cmd);
     if ((strlen(cmd) + 3) > sizeof(buf))
         return 0;
     sprintf(buf,"%s\r\n",cmd);
-    if (net_write(nControl->handle,buf,strlen(buf)) <= 0)
+    if (netbuf_write (nControl,buf,strlen(buf)) <= 0)
     {
 	/* perror("write"); */
 	return 0;
@@ -600,6 +672,7 @@ GLOBALDEF int FtpLogin(const char *user, const char *pass, netbuf *nControl)
     if (((strlen(user) + 7) > sizeof(tempbuf)) ||
         ((strlen(pass) + 7) > sizeof(tempbuf)))
         return 0;
+
     sprintf(tempbuf,"USER %s",user);
     if (!FtpSendCmd(tempbuf,'3',nControl))
     {
@@ -611,7 +684,7 @@ GLOBALDEF int FtpLogin(const char *user, const char *pass, netbuf *nControl)
     return FtpSendCmd(tempbuf,'2',nControl);
 }
 
-/* 
+/*
  * FtpAuthorize - authorize FTP proxy to send outgoing data; requires
  * user interaction with cryptokey
  */
@@ -859,7 +932,7 @@ static int FtpAcceptConnection(netbuf *nData, netbuf *nControl)
 	    rv = 0;
 	}
     }
-    return rv;	
+    return rv;
 }
 
 /*
@@ -929,6 +1002,18 @@ GLOBALDEF int FtpAccess(const char *path, int typ, int mode, netbuf *nControl,
 	    return 0;
 	}
     }
+#ifdef USE_SSL
+   if (nControl->ssl_con != NULL)
+   {
+	if (-1 == setup_data_connection_ssl (*nData))
+        {
+	    FtpClose (*nData);
+	    *nData = NULL;
+	    return 0;
+        }
+    }
+#endif
+
     return 1;
 }
 
@@ -945,7 +1030,7 @@ GLOBALDEF int FtpRead(void *buf, int max, netbuf *nData)
     else
     {
         socket_wait(nData);
-        i = net_read(nData->handle, buf, max);
+        i = netbuf_read (nData, buf, max);
     }
     nData->xfered += i;
     if (nData->idlecb && nData->cbbytes)
@@ -966,6 +1051,7 @@ GLOBALDEF int FtpRead(void *buf, int max, netbuf *nData)
 GLOBALDEF int FtpWrite(void *buf, int len, netbuf *nData)
 {
     int i;
+
     if (nData->dir != FTPLIB_WRITE)
 	return 0;
     if (nData->buf)
@@ -973,7 +1059,7 @@ GLOBALDEF int FtpWrite(void *buf, int len, netbuf *nData)
     else
     {
         socket_wait(nData);
-        i = net_write(nData->handle, buf, len);
+        i = netbuf_write (nData, buf, len);
     }
     nData->xfered += i;
     if (nData->idlecb && nData->cbbytes)
@@ -1003,6 +1089,12 @@ GLOBALDEF int FtpClose(netbuf *nData)
 	return 0;
     if (nData->buf)
     	free(nData->buf);
+
+#ifdef USE_SSL
+   if (nData->ssl_con != NULL)
+     SSL_free (nData->ssl_con);
+#endif
+
     shutdown(nData->handle,2);
     net_close(nData->handle);
     ctrl = nData->ctrl;
@@ -1027,6 +1119,12 @@ GLOBALDEF int FtpClose2(netbuf *nData)
       return 0;
     if (nData->buf)
       free(nData->buf);
+
+#ifdef USE_SSL
+   if (nData->ssl_con != NULL)
+     SSL_free (nData->ssl_con);
+#endif
+
     shutdown(nData->handle,2);
     net_close(nData->handle);
     ctrl = nData->ctrl;
@@ -1044,7 +1142,7 @@ GLOBALDEF int FtpClose2(netbuf *nData)
 GLOBALDEF int FtpSite(const char *cmd, netbuf *nControl)
 {
     char buf[256];
-    
+
     if ((strlen(cmd) + 7) > sizeof(buf))
         return 0;
     sprintf(buf,"SITE %s",cmd);
@@ -1203,6 +1301,9 @@ static int FtpXfer(const char *localfile, const char *path,
         while ((l = fread(dbuf, 1, FTPLIB_BUFSIZ, local)) > 0){
 	    if ((c = FtpWrite(dbuf, l, nData)) < l){
 		printf("short write: passed %d, wrote %d\n", l, c);
+		/* we should handle short writes more gracefully */
+		FtpClose(nData);
+		return 0;
             }
             if(opt_silent){
                 continue;
@@ -1294,7 +1395,7 @@ GLOBALDEF int FtpSize(const char *path, int *size, char mode, netbuf *nControl)
 	    *size = sz;
 	else
 	    rv = 0;
-    }   
+    }
     return rv;
 }
 
@@ -1407,8 +1508,179 @@ GLOBALDEF void FtpQuit(netbuf *nControl)
     if (nControl->dir != FTPLIB_CONTROL)
 	return;
     FtpSendCmd("QUIT",'2',nControl);
+#ifdef USE_SSL
+    if (nControl->ssl_con != NULL)
+    {
+	/* SSL_free (nControl->ssl_con); -- not clear if this affects ->handle */
+    }
+#endif
     net_close(nControl->handle);
+    free(nControl->hostname);
     free(nControl->buf);
     free(nControl);
 }
 
+#ifdef USE_SSL
+static SSL_CTX *SSL_Context;
+
+static int print_ssl_err_message (char *fun, SSL *con, int status)
+{
+    char *err;
+
+    switch (SSL_get_error (con, status))
+	{
+	case SSL_ERROR_ZERO_RETURN:
+	    return 0;
+
+	case SSL_ERROR_WANT_READ:
+	    err = "SSL_ERROR_WANT_READ";
+	    break;
+
+	case SSL_ERROR_WANT_WRITE:
+	    err = "SSL_ERROR_WANT_WRITE";
+	    break;
+
+	case SSL_ERROR_WANT_CONNECT:
+	    err = "SSL_ERROR_WANT_CONNECT";
+	    break;
+#if 0
+	case SSL_ERROR_WANT_ACCEPT:
+	    err = "SSL_ERROR_WANT_ACCEPT";
+	    break;
+#endif
+	case SSL_ERROR_WANT_X509_LOOKUP:
+	    err = "SSL_ERROR_WANT_X509_LOOKUP";
+	    break;
+
+	case SSL_ERROR_SYSCALL:
+	    err = "SSL_ERROR_SYSCALL";
+	    break;
+
+	case SSL_ERROR_SSL:
+	    err = "SSL_ERROR_SSL";
+	    break;
+
+	default:
+	    err = "SSL_ERROR_UNKNOWN";
+	    break;
+	}
+
+    fprintf (stderr, "%s failed: %s (status=%d)\n", fun, err, status);
+    return -1;
+}
+
+static int authorize_ssl (netbuf *c)
+{
+    int status;
+
+    if (-1 == initialize_ssl ())
+	return -1;
+
+    if (0 == FtpSendCmd ("AUTH SSL", 127, c)) {
+	    fprintf (stderr, "SSL not supported.  Secure connection failed\n");
+	    return -1;
+    }
+
+    c->ssl_con = (SSL *)SSL_new (SSL_Context);
+    if (c->ssl_con == NULL)
+	{
+	    fprintf (stderr, "**** SSL_new failed\n");
+	    return -1;
+	}
+
+    SSL_set_fd(c->ssl_con, c->handle);
+
+    /* Certificate processing goes here */
+    if(host_use_SSL_verify[host_number])
+    {
+	if(! SSL_CTX_load_verify_locations(SSL_Context, host_use_SSL_cafile[host_number], NULL))
+	{
+	    /* Handle failed load here */
+	    fprintf (stderr, "**** Could not load certificates\n");
+	    return -1;
+	}
+    }
+
+    status = SSL_connect (c->ssl_con);
+    if (status < 0)
+	return print_ssl_err_message ("SSL_connect", c->ssl_con, status);
+
+    fprintf(stdout,"           : [SSL Cipher %s]\n", SSL_get_cipher(c->ssl_con));
+    fflush(stdout);
+
+    if(host_use_SSL_verify[host_number])
+	{
+	    X509 *ssl_client_cert = NULL;
+	
+	    ssl_client_cert = SSL_get_peer_certificate(c->ssl_con);
+	
+	    if(ssl_client_cert)
+		{
+		    long verifyresult;
+		    X509 *peer;
+		    char peer_CN[256];
+	     
+		    verifyresult = SSL_get_verify_result(c->ssl_con);
+		    if(verifyresult != X509_V_OK) {
+			fprintf (stderr, "**** Certificate Verify Failed error=%ld\n", verifyresult); 
+			return -1;
+		    }
+		    /*Check the common name*/
+		    peer = SSL_get_peer_certificate(c->ssl_con);
+		    X509_NAME_get_text_by_NID(X509_get_subject_name(peer), NID_commonName, peer_CN, 256);
+		    if (strcasecmp(peer_CN, c->hostname)) {
+			fprintf (stderr, "**** Certificate name does not match:%s\n", peer_CN); 
+			return -1;
+		    }
+		    X509_free(ssl_client_cert);
+		}
+	    else
+		fprintf(stderr, "**** There is no client certificate\n");
+		
+	}
+    return 0;
+}
+
+int initialize_ssl (void)
+{
+    if (SSL_Context != NULL)
+	return 0;			       /* already inited */
+   
+    (void) SSL_load_error_strings();
+    (void) SSLeay_add_ssl_algorithms();
+
+    if (NULL == (SSL_Context = (SSL_CTX *)SSL_CTX_new(SSLv23_method())))
+	{
+	    fprintf (stderr, "SSL_CTX_new failed\n");
+	    return -1;
+	}
+    SSL_CTX_set_options (SSL_Context, SSL_OP_NO_SSLv2);
+
+    return 0;
+}
+
+static int setup_data_connection_ssl (netbuf *d)
+{
+    int status;
+
+    d->ssl_con = (SSL *)SSL_new (SSL_Context);
+    if (d->ssl_con == NULL)
+	{
+	    fprintf (stderr, "**** SSL_new failed\n");
+	    return -1;
+	}
+    
+    SSL_set_fd(d->ssl_con, d->handle);
+#if 1
+    if ((d->ctrl != NULL) && (d->ctrl->ssl_con != NULL))
+	SSL_copy_session_id (d->ssl_con, d->ctrl->ssl_con);
+#endif
+
+    status = SSL_connect (d->ssl_con);
+    if (status < 0)
+	return print_ssl_err_message ("SSL_connect", d->ssl_con, status);
+
+    return 0;
+}
+#endif /* USE_SSL */
+
